local _, addon_table = ...
local EnhancedModelScene = addon_table.self

local Commands = {}
EnhancedModelScene.Commands = Commands

--[[
need flag for how command should be synced
as-is, revise, don't-sync
]]


Commands.commands = {}

function Commands:Dispatch(input)
	local command, resume_at = EnhancedModelScene:GetArgs(input)
	
	--print("checking for command:", command)
	
	local func = self.commands[command]
	
	if func then
		func(command, input, resume_at)
	else
		-- try the old command block
		
		local no_match

		local cmd, remainder = input:match("(%S+)%s*(.*)")
		cmd = cmd:lower()
		
		local self = EnhancedModelScene
		
		--print("trying old command block")
		print(cmd, remainder)
		
		if cmd == "new" or cmd == "newactor" then
			self:NewActor(remainder)
			
			print("changing active model from", self.active_model, "to", self:GetNumActors())
			self.active_model = self:GetNumActors()
		elseif cmd == "set" or cmd == "change" then
			local index, target = strsplit(" ", remainder)
			self:SetActorTo(self:GetActorAtIndex(tonumber(index)), target)
		elseif cmd == "setfile" then
			local index, target = strsplit(" ", remainder)
			-- FIXME: warn and require SV flag, also could check a database
			local actor = self:GetActorAtIndex(tonumber(index))
			actor:SetModelByFileID(tonumber(target))
			actor.who = nil
		elseif cmd == "store" then
			-- store to a variable
		elseif cmd == "anim" or cmd == "animation" then
			-- FIXME: arguments
			local index, anim, variation, speed, offset = strsplit(" ", remainder)
			index, anim, variation = tonumber(index), tonumber(anim) or 0, tonumber(variation) or 0
			offset, speed = tonumber(offset) or 0, tonumber(speed) or 1
			
			local actor = self:GetActorAtIndex(index)
			actor:SetAnimation(anim, variation, speed, offset)
			actor.animation = anim
			actor.offset = offset
			actor.speed = speed
			actor.variation = variation
			
			if self.log then
				tinsert(self.log, { GetTime(), "SetAnimation", index, anim, variation, speed, offset} )
			end
		elseif cmd == "scale" then
			local index, scale = strsplit(" ", remainder)
			
			local actor = self:GetActorAtIndex(tonumber(index))
			
			if self.fix_scale then
				local old_scale = actor:GetScale()
				local x, y, z = actor:GetPosition()
				local k = old_scale / scale
				x, y, z = k*x, k*y, k*z
				actor:SetPosition(x, y, z)
			end
			
			actor:SetScale(tonumber(scale))
			
			if self.log then
				tinsert(self.log, { GetTime(), "SetScale", index, scale} )
			end
		elseif cmd == "move" or cmd == "adjust" or cmd == "active" or cmd == "setcurrent" then
			-- set the model at the index for individual adjustment
			self.active_model = tonumber(remainder)
		elseif cmd == "getcurrent" then
			print(self.active_model)
		elseif cmd == "hide" then
			local index = tonumber(remainder)
			self:GetActorAtIndex(index):Hide()
		elseif cmd == "hideall" then
			for i = 1, self:GetNumActors() do
				self:GetActorAtIndex(i):Hide()
			end
		elseif cmd == "show" then
			local index = tonumber(remainder)
			print(index)
			self:ShowModel(index)
		elseif cmd == "showall" then
			for i = 1, self:GetNumActors() do
				self:GetActorAtIndex(i):Show()
			end
		elseif cmd == "setglobal" then
			local index, name = strsplit(" ", remainder)
			if not name then name = "a" .. index end
			index = tonumber(index)
			
			print(index, name)
			
			_G[name] = self:GetActorAtIndex(index)
		elseif cmd == "backdrop" or cmd == "bd" then
			self:ToggleBackdrop()
		elseif cmd == "fov" then
			local degrees = tonumber(remainder)
			self:SetCameraFieldOfView(rad(degrees))
		elseif cmd == "undress" then
			self:GetActorAtIndex(tonumber(remainder)):Undress()
		elseif cmd == "equip" then
			local index, item = strsplit(" ", remainder)
			self:GetActorAtIndex(tonumber(index)):TryOn("item:" .. item)
		elseif cmd == "flash" then
			local index = tonumber(remainder)
			local model = self:GetActorAtIndex(index)
			if model:IsShown() then
				model:Hide()
				C_Timer.After(1, function() model:Show() end)
			else
				model:Show()
				C_Timer.After(1, function() model:Hide() end)
			end
		elseif cmd == "pos" or cmd == "position" or cmd == "xyz" then
			local index, x, y, z = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetScaledPosition(x, y, z)
		elseif cmd == "rawpos" or cmd == "rawposition" then
			local index, x, y, z = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetPosition(x, y, z)
		elseif cmd == "getpos" then
			local index = tonumber(remainder)
			local actor = self:GetActorAtIndex(index)
			print(actor:GetPosition())
			
			if self.fix_scale then
				local s = actor:GetScale()
				local x, y, z = actor:GetPosition()
				print(x*s, y*s, z*s)
			end
		elseif cmd == "x" then
			local index, v = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			local x, y, z = actor:GetScaledPosition()
			x = v
			actor:SetScaledPosition(x, y, z)
		elseif cmd == "y" then
			local index, v = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			local x, y, z = actor:GetScaledPosition()
			y = v
			actor:SetScaledPosition(x, y, z)
		elseif cmd == "z" then
			local index, v = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			local x, y, z = actor:GetScaledPosition()
			z = v
			actor:SetScaledPosition(x, y, z)
		elseif cmd == "+x" then
			-- FIXME: adjust nudges by scale
			local index, v = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			local x, y, z = actor:GetScaledPosition()
			x = x + v
			actor:SetScaledPosition(x, y, z)
		elseif cmd == "+y" then
			local index, v = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			local x, y, z = actor:GetScaledPosition()
			y = y + v
			actor:SetScaledPosition(x, y, z)
		elseif cmd == "+z" then
			local index, v = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			local x, y, z = actor:GetScaledPosition()
			z = z + v
			actor:SetScaledPosition(x, y, z)
		elseif cmd == "yaw" then
			local index, degrees = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetYaw(rad(degrees))
		elseif cmd == "pitch" then
			local index, degrees = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetPitch(rad(degrees))
		elseif cmd == "roll" then
			local index, degrees = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetRoll(rad(degrees))
		elseif cmd == "ryaw" or cmd == "yawr" then
			local index, radians = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetYaw(radians)
		elseif cmd == "rpitch" or cmd == "pitchr" then
			local index, radians = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetPitch(radians)
		elseif cmd == "rroll" or cmd == "rollr" then
			local index, radians = strsplit(" ", remainder)
			local model = self:GetActorAtIndex(tonumber(index))
			model:SetRoll(radians)
		elseif cmd == "orientation" or cmd == "ori" or 
		cmd == "ypr" or cmd == "yprd" or cmd == "yprdeg" or cmd == "yprdegrees" then
			local index, yaw, pitch, roll = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			actor:SetYaw(rad(tonumber(yaw)))
			actor:SetPitch(rad(tonumber(pitch)))
			actor:SetRoll(rad(tonumber(roll)))
		elseif cmd == "yprr" or cmd == "yprrad" or cmd == "yprradians" then
			local index, yaw, pitch, roll = strsplit(" ", remainder)
			local actor = self:GetActorAtIndex(tonumber(index))
			actor:SetYaw(tonumber(yaw))
			actor:SetPitch(tonumber(pitch))
			actor:SetRoll(tonumber(roll))
		elseif cmd == "whois" then
			local index = tonumber(remainder)
			print(self:GetActorAtIndex(index).who)
		elseif cmd == "input" then
			if not self.input_frame then
				self:MakeInputFrame()
			end
			C_Timer.After(1, function() self.input_frame:Show() end)
		elseif cmd == "relpos" then
		elseif cmd == "copypos" then
		elseif cmd == "lockmodel" then
			local a = self:GetActorAtIndex(tonumber(remainder))
			a.lock_model = true
		elseif cmd == "unlockmodel" then
			local a = self:GetActorAtIndex(tonumber(remainder))
			a.lock_model = false
		elseif cmd == "alpha" then
			local index, value = strsplit(" ", remainder)
			self:GetActorAtIndex(tonumber(index)):SetAlpha(tonumber(value) or 1)	
		elseif cmd == "desat" then
			local index, value = strsplit(" ", remainder)
			self:GetActorAtIndex(tonumber(index)):SetDesaturation(tonumber(value) or 1)
		elseif cmd == "getfile" then
			print(self:GetActorAtIndex(tonumber(remainder)):GetModelFileID())
		elseif cmd == "rotatemodel" or cmd == "rotate" then
			-- rotate actor around its internal axis
			local index, axis, angle = strsplit(" ", remainder)
			-- FIXME: might need to adjust angle value if the actor is scaled
			
			local a = self:GetActorAtIndex(tonumber(index))
			
			local initial = a:GetRotationMatrix()
			local rotation = self.Matrix.AngleAxis(rad(tonumber(angle)), initial:GetBasisVector(tonumber(axis)))
			
			a:SetOrientation(self.Matrix.hmult(rotation, initial))
		elseif cmd == "rotaterel" or cmd == "rotateref" then
			-- rotate actor around a different actor's axis
			local target, reference, axis, angle = strsplit(" ", remainder)
			target = self:GetActorAtIndex(tonumber(target))
			reference = self:GetActorAtIndex(tonumber(reference))

			-- FIXME: might need to adjust angle value if the actor is scaled
			
			local initial = target:GetRotationMatrix()
			local rotation = self.Matrix.AngleAxis(rad(tonumber(angle)), reference:GetBasisVector(tonumber(axis)))
			
			target:SetOrientation(self.Matrix.hmult(rotation, initial))
		elseif cmd == "slide" then
			-- shift a model along an internal axis
			local index, axis, distance = strsplit(" ", remainder)
			local a = self:GetActorAtIndex(tonumber(index))
			a:SlideOnAxis(tonumber(axis), tonumber(distance))
		elseif cmd == "slideother" or cmd == "slideo" or cmd == "slideby" then
			local x, y, z = reference:GetRotationMatrix():GetBasisVector(tonumber(axis))
			target:Slide(x, y, z, tonumber(distance))
		elseif cmd == "showtarget" then
			if not self.camera_target then
				self.camera_target = self:NewActor(34715)
				self.camera_target:SetScaledPosition(self:GetCameraTarget())
			end
		elseif cmd == "listsessions" then
			self:PrintSessionSpans()
		elseif cmd == "loadlastsession" or cmd == "lastsession" then
			self:LoadLastSession()
		elseif cmd == "loadsession" then
			local index = tonumber(remainder)
			self:LoadSession(index)
		elseif cmd == "rebuild" then
			local index = tonumber(remainder) or #self.sv.sessions
			self:RebuildActorsFromScene(self.sv.sessions[index])
		elseif cmd == "continue" then
			self:ContinueLastSession()
		elseif cmd == "name" then
			local index, name = strsplit(" ", remainder)
			self:GetActorAtIndex(tonumber(index)).who = name
		elseif cmd == "allowsync" then
			self.allow_sync = true
		elseif cmd == "stopsync" then
			self.allow_sync = false
		elseif cmd == "fullsync" then
			self.sync = true
			self.allow_sync = true
			self.group_sync = true
		elseif cmd == "getscale" then
			print(self:GetActorAtIndex(tonumber(remainder)):GetScale())
		elseif cmd == "copy" then
			local source, target = strsplit(" ", remainder)
			source, target = tonumber(source), tonumber(target)
			self.active_model = target
			source, target = self:GetActorAtIndex(source), self:GetActorAtIndex(target)
			target:SetScale(source:GetScale())
			target:SetPosition(source:GetPosition())
			target:SetYPR(source:GetYPR())
			target.who = source.who
		elseif cmd == "link" then
			-- TODO: support list of models to link to first
			local leader, follower = strsplit(" ", remainder)
			self:LinkActorTo(tonumber(follower), tonumber(leader))
		elseif cmd == "searchfiles" then
			self:ScrollMatchingFiles(remainder)
		elseif cmd == "walk" or cmd == "walkspeed" then
			local index, speed = strsplit(" ", remainder)
			index, speed = tonumber(index), tonumber(speed)
			local actor = self:GetActorAtIndex(index)
			actor.move_speed = speed
			actor:PickAnimation()
		elseif cmd == "turn" or cmd == "turnspeed" then
			local index, speed = strsplit(" ", remainder)
			index, speed = tonumber(index), tonumber(speed)
			local actor = self:GetActorAtIndex(index)
			actor.turn_speed = speed			
			actor:PickAnimation()
		elseif cmd == "immersion" then
			self.Immersion:SetShown(not self.Immersion:IsShown())
		elseif cmd == "resetcamera" then
			self:ResetCamera()
		else
			no_match = true
		end
		
		if no_match then
			print("unrecognized command")
		end
	end
end
